//Martin Rheinländer, SS 05

#include "MWDrawArea.h"

MWDrawArea::MWDrawArea(QWidget *parent, const char *name ):QWidget(parent, name){
	DrawMode   = 1;
	PointSize  = 5;
	GridSnapON = true;
	setPalette(QPalette(QColor(255,255,255)));
	setCoordRange(-1.0, 1.0, -1.0, 1.0);
	setGridSpacing(0.1, 0.1);
	sx = 1.0/(xmax-xmin);
	sy = 1.0/(ymax-ymin);
}

MWDrawArea::~MWDrawArea(){}

void MWDrawArea::setDrawMode(int m){
	DrawMode = m;
}

void MWDrawArea::setPointSize(int p) {
	PointSize = p;
	this->repaint();
}

void MWDrawArea::setGridSnap(bool b) {
	GridSnapON = b;
}

void MWDrawArea::setCoordRange(double x_min, double x_max, double y_min, double y_max) {
	xmin = x_min; xmax = x_max;
	ymin = y_min; ymax = y_max;
}

void MWDrawArea::setGridSpacing(double dx_init, double dy_init) {
	dx = dx_init;
	dy = dy_init;
}

QPoint MWDrawArea::getPixel(const MWPoint& Point){         
	int PixelX =        (int)(     round((Point.getx()-xmin)/(xmax-xmin)*width()));
	int PixelY = (int) (height() - round((Point.gety()-ymin)/(ymax-ymin)*height()));	
        return QPoint(PixelX,PixelY);
}	

int MWDrawArea::getPixelX(const MWPoint& Point){
	int px = (int) (round((Point.getx()-xmin)/(xmax-xmin)*width()));
	return px;
}

int MWDrawArea::getPixelY(const MWPoint& Point){
	int py = (int) (height() - round((Point.gety()-ymin)/(ymax-ymin)*height()));
	return py;
}

MWPoint MWDrawArea::getCoord(const QPoint& Pixel){
	double x = xmin + ( -1     +Pixel.x())*(xmax-xmin)/width(); 
	double y = ymin + (height()-Pixel.y())*(ymax-ymin)/height();
	return MWPoint(x,y);
}

void MWDrawArea::clearConvexHull(){
	UpperConvexHull.emptyout();
	LowerConvexHull.emptyout();
	ConvexHull.emptyout();
	currentIndex =0;
}

void MWDrawArea::undo(){
        PointCloud.eraseTopMost(0);
	clearConvexHull();
	this->repaint();
}

void MWDrawArea::paintEvent(QPaintEvent*) {
        std::cout << "MWDrawArea::paintEvent: " << PointCloud.getCardinal() << std::endl;
	QPainter Painter(this);
	QBrush   Brush(black,SolidPattern);

	Painter.setPen(QPen(black,0,SolidLine));
	Painter.setBrush(Brush);
	 
	for(int i=0; i<PointCloud.getCardinal(); i++)
		Painter.drawEllipse(getPixelX(PointCloud(i))-PointSize/2,
		                    getPixelY(PointCloud(i))-PointSize/2, 
				    PointSize,PointSize);
	
	if(ConvexHull.getCardinal()>0){
	        Painter.setPen(QPen(black,0,SolidLine));
        	for(int i=0; i<ConvexHull.getCardinal()-1; i++)
			Painter.drawLine(getPixel(ConvexHull(i)), getPixel(ConvexHull(i+1)) );
		Painter.drawLine(getPixel(ConvexHull.getTopMost(0)), getPixel(ConvexHull(0)) );	
        }
	
	Painter.setPen(QPen(red,0,SolidLine));
        for(int i=0; i<UpperConvexHull.getCardinal()-1; i++)
	
		Painter.drawLine(getPixel(UpperConvexHull(i)), getPixel(UpperConvexHull(i+1)) );	

	Painter.setPen(QPen(green,0,SolidLine));
        for(int i=0; i<LowerConvexHull.getCardinal()-1; i++)
		Painter.drawLine(getPixel(LowerConvexHull(i)), getPixel(LowerConvexHull(i+1)) );
}

void MWDrawArea::mouseReleaseEvent(QMouseEvent* Event){
        std::cout << "MWDrawArea::mouseReleaseEvent: " << PointCloud.getCardinal() << std::endl;
	if(Event->button()==LeftButton){
        MWPoint Point;
	switch(DrawMode){
		case 0: Event->ignore();
			break;
		case 1: Event->accept();
			if(GridSnapON){
				Point = getCoord(Event->pos());
				Point[1] = round(Point[1]/dx) * dx;
				Point[2] = round(Point[2]/dy) * dy;
			}else   Point = getCoord(Event->pos());
			PointCloud.append(Point);
			clearConvexHull();
			this->repaint();
			std::cout << "Point added: " << Point[1] << ", " << Point[2] << std::endl;
			std::cout << "MWDrawArea::mouseReleaseEvent: " << PointCloud.getCardinal() << std::endl;
			break;
	}//switch
	}//if
	else if(Event->button()==RightButton){
		if(DrawMode==1) DrawMode=0;
		else DrawMode = 1; 
	}//else if
	else Event->ignore();
}


void MWDrawArea::keyPressEvent(QKeyEvent* Event){
	if     (Event->key()==QKeyEvent::Key_Backspace){
		Event->accept();
		undo();
	}
	else if(Event->key()==QKeyEvent::Key_F){
		findConvexHull();
		this->repaint();
	}
	else if(Event->key()==QKeyEvent::Key_C){
	        constructConvexHull();
		this->repaint();
	}
	else Event->ignore();	
}

void MWDrawArea::findConvexHull(){
        List<MWPoint> LowConvexHull;
	List<MWPoint> UppConvexHull;
	PointCloud.sort();
	//PointCloud.print();
	//find upper convex hull
	UppConvexHull.append(PointCloud(0));
	UppConvexHull.append(PointCloud(1));
	for(int i=2; i<PointCloud.getCardinal(); i++){
		UppConvexHull.append(PointCloud(i));
		while(UppConvexHull.getCardinal()>2){
        		if(isStrictLeftTurn(UppConvexHull.getTopMost(2),
			                    UppConvexHull.getTopMost(1),
					    UppConvexHull.getTopMost(0) )) 
				UppConvexHull.eraseTopMost(1);
			else break;
		}//while
	}//for
	//UpperConvexHull->print();
	//find lower convex hull
	LowConvexHull.append(PointCloud(0));
	LowConvexHull.append(PointCloud(1));
	for(int i=2; i<PointCloud.getCardinal(); i++){
		LowConvexHull.append(PointCloud(i));
		while(LowConvexHull.getCardinal()>2){
        		if(isStrictRightTurn(LowConvexHull.getTopMost(2),
			                     LowConvexHull.getTopMost(1),
					     LowConvexHull.getTopMost(0) )) 
				LowConvexHull.eraseTopMost(1);
			else break;
		}//while
	}//for
	//LowerConvexHull->print();
	LowConvexHull.eraseBotMost(0); 
	LowConvexHull.eraseTopMost(0);
	
	for(int i=0; i<UppConvexHull.getCardinal(); i++)
		ConvexHull.append(UppConvexHull(i));
		
	for(int i=LowConvexHull.getCardinal()-1; i>=0; i--)
		ConvexHull.append(LowConvexHull(i));
}//findConvexHull


void MWDrawArea::constructConvexHull(){
	PointCloud.sort();
	//find upper convex hull
	if(currentIndex==0){
		UpperConvexHull.append(PointCloud(0));
		UpperConvexHull.append(PointCloud(1));
		LowerConvexHull.append(PointCloud(0));
	        LowerConvexHull.append(PointCloud(1));
		currentIndex=2;
	}else if(currentIndex < PointCloud.getCardinal()){
		UpperConvexHull.append(PointCloud(currentIndex));
		while(UpperConvexHull.getCardinal()>2){
        		if(isStrictLeftTurn(UpperConvexHull.getTopMost(2),
			                    UpperConvexHull.getTopMost(1),
					    UpperConvexHull.getTopMost(0) )) 
				UpperConvexHull.eraseTopMost(1);
			else break;
		}	
		LowerConvexHull.append(PointCloud(currentIndex));
		while(LowerConvexHull.getCardinal()>2){
        		if(isStrictRightTurn(LowerConvexHull.getTopMost(2),
			                     LowerConvexHull.getTopMost(1),
					     LowerConvexHull.getTopMost(0) )) 
				LowerConvexHull.eraseTopMost(1);
			else break;
		}	
		currentIndex++;	
	}else{;}
}//constructConvexHull



